#include "ImgToHilbert.h"
#include <vector>
#include <algorithm>

ImgToHilbert::ImgToHilbert()
{
	State = 0;

	int i;
	Power2[0] = 1;
	for (i = 1; i < 32; i++)
		Power2[i] = 2 * Power2[i - 1];

	W = H = 0;
}

void ImgToHilbert::SetImage(int wi, int he, vector<vector<int>> Img)
{
	I = Img;
	W = wi;
	H = he;

	DetermineOrder();
}

void ImgToHilbert::DetermineOrder()
{
	int ox, oy;

	ox = oy = 0;
	while (W > Power2[ox])
		ox++;

	while (H > Power2[oy])
		oy++;

	Order = ox;
	if (Order < oy)
		Order = oy;
}


void ImgToHilbert::ToHilbert()
{
/*	W = 16;
	H = 16;

	State = 0;
	Order = 4;
	int x, y;
	x = 0; 
	y = 15;
	ConvertToHIndex(x, y);
	int hi1 = (int)GenerateHIndex();
	*/

	int hi;
	for (int x = 0; x < W; x++)
		for (int y = 0; y < H; y++)
		{
			State = 0;
			ConvertToHIndex(x, y);

			if ((x == 256) && (y == 0))
			{
				int stop = 0;
			}

			hi = (int)GenerateHIndex();	
			HI.push_back(hi);
		
		}

	sort(HI.begin(), HI.end());
	HPixels.clear();



	int x, y;
	int c;
	for (int i = 0; i < HI.size(); i++)
	{
		State = 0;
		CovertToXY(HI[i], x, y);
		c = I[x][y];
		HPixels.push_back(c);
	}

}


void ImgToHilbert::ConvertToHIndex(int x, int y)
{
	for (int i = 0; i < 32; i++)
		HInputX[i] = HInputY[i] = 0;


	for (int i = Order - 1; i >= 0; i--)
	{
		HInputX[i] = (Power2[i] & x) >> i;
		HInputY[i] = (Power2[i] & y) >> i;
	}
}


unsigned long ImgToHilbert::GenerateHIndex()
{
	HIndex = 0;
	for (int i = Order - 1; i >= 0; i--)
	{
		switch (State)
		{
		case 0: SolveState0(HInputX[i], HInputY[i]); 	break;
		case 1: SolveState1(HInputX[i], HInputY[i]); 	break;
		case 2: SolveState2(HInputX[i], HInputY[i]); 	break;
		case 3: SolveState3(HInputX[i], HInputY[i]); 	break;
		}
	}
	return HIndex;
}


void ImgToHilbert::SolveState0(int i, int j)
{
	if ((i == 0) && (j == 0))
	{
		State = 1;
		AddToHIndex(0, 0);
	}

	if ((i == 1) && (j == 0))
	{
		State = 2;
		AddToHIndex(1, 1);
	}

	if ((i == 0) && (j == 1))	AddToHIndex(0, 1);
	if ((i == 1) && (j == 1))	AddToHIndex(1, 0);
}

void ImgToHilbert::SolveState1(int i, int j)
{
	if ((i == 0) && (j == 0))
	{
		State = 0;
		AddToHIndex(0, 0);
	}

	if ((i == 0) && (j == 1))
	{
		State = 3;
		AddToHIndex(1, 1);
	}

	if ((i == 1) && (j == 1))	AddToHIndex(1, 0);
	if ((i == 1) && (j == 0))	AddToHIndex(0, 1);
}

void ImgToHilbert::SolveState2(int i, int j)
{
	if ((i == 1) && (j == 0))
	{
		State = 0;
		AddToHIndex(1, 1);
	}

	if ((i == 1) && (j == 1))
	{
		State = 3;
		AddToHIndex(0, 0);
	}

	if ((i == 0) && (j == 1))	AddToHIndex(0, 1);
	if ((i == 0) && (j == 0))	AddToHIndex(1, 0);
}

void ImgToHilbert::SolveState3(int i, int j)
{
	if ((i == 0) && (j == 1))
	{
		State = 1;
		AddToHIndex(1, 1);
	}

	if ((i == 1) && (j == 1))
	{
		State = 2;
		AddToHIndex(0, 0);
	}

	if ((i == 0) && (j == 0))	AddToHIndex(1, 0);
	if ((i == 1) && (j == 0))	AddToHIndex(0, 1);
}


void ImgToHilbert::AddToHIndex(int x, int y)
{
	HIndex = HIndex << 1;
	HIndex += x;
	HIndex = HIndex << 1;
	HIndex += y;

}




void ImgToHilbert::CovertToXY(int index, int& x, int& y)
{
//	index = 0;
//	Order = 4;


	unsigned long h = (unsigned long) index;

	int j;
	vector<unsigned int> stack;
	unsigned int a, b;
	int bx, by;
	unsigned int qx, qy; 


	for (j = 0; j < Order; j++)
	{
		a = h & 1;
		stack.insert(stack.begin(), a);
		h = h >> 1;

		b = h & 1;
		stack.insert(stack.begin(), b);
		h = h >> 1;
	}


	State = 0;
	qx = qy = 0;

	for (j = 0; j < 2 * Order; j += 2)
	{
		a = stack[j];
		b = stack[j + 1];

		qx = qx << 1;
		qy = qy << 1;
		switch (State)
		{
		case 0: State = DecodeState0(a, b, bx, by); break;
		case 1: State = DecodeState1(a, b, bx, by); break;
		case 2: State = DecodeState2(a, b, bx, by); break;
		case 3: State = DecodeState3(a, b, bx, by); 
		}
		qx += bx;
		qy += by;

	}
	stack.clear();
	x = qx;
	y = qy;
}


int ImgToHilbert::DecodeState0(unsigned int a, unsigned int b, int& x, int& y)
{

	if ((a == 0) && (b == 0))
	{
		x = 0;		y = 0;
		return 1;
	}

	if ((a == 0) && (b == 1))
	{
		x = 0;		y = 1;
		return 0;
	}

	if ((a == 1) && (b == 0))
	{
		x = 1;		y = 1;
		return 0;
	}

	if ((a == 1) && (b == 1))
	{
		x = 1;		y = 0;
		return 2;
	}
	return 0;
}


int ImgToHilbert::DecodeState1(unsigned int a, unsigned int b, int& x, int& y)
{

	if ((a == 0) && (b == 0))
	{
		x = 0;		y = 0;
		return 0;
	}

	if ((a == 0) && (b == 1))
	{
		x = 1;		y = 0;
		return 1;
	}

	if ((a == 1) && (b == 0))
	{
		x = 1;		y = 1;
		return 1;
	}

	if ((a == 1) && (b == 1))
	{
		x = 0;		y = 1;
		return 3;
	}
	return 0;
}

int ImgToHilbert::DecodeState2(unsigned int a, unsigned int b, int& x, int& y)
{

	if ((a == 0) && (b == 0))
	{
		x = 1;		y = 1;
		return 3;
	}

	if ((a == 0) && (b == 1))
	{
		x = 0;		y = 1;
		return 2;
	}

	if ((a == 1) && (b == 0))
	{
		x = 0;		y = 0;
		return 2;
	}

	if ((a == 1) && (b == 1))
	{
		x = 1;		y = 0;
		return 0;
	}
	return 0;
}


int  ImgToHilbert::DecodeState3(unsigned int a, unsigned int b, int& x, int& y)
{

	if ((a == 0) && (b == 0))
	{
		x = 1;		y = 1;
		return 2;
	}

	if ((a == 0) && (b == 1))
	{
		x = 1;		y = 0;
		return 3;
	}

	if ((a == 1) && (b == 0))
	{
		x = 0;		y = 0;
		return 3;
	}

	if ((a == 1) && (b == 1))
	{
		x = 0;		y = 1;
		return 1;
	}
	return 0;
}




//// De
void ImgToHilbert::DeSetImageResolution(int wi, int he)
{
	W = wi;
	H = he;

	DetermineOrder();
}

void ImgToHilbert::DeReturnMatrix(vector<int>& Y, vector<vector<int>>& M)
{
	int hi;
	int i = 0;
	vector<int> Q; // vector to match position of Hilbert indices in the actual vector;
	
	
	for (int x = 0; x < W; x++)
		for (int y = 0; y < H; y++)
		{
			State = 0;
			ConvertToHIndex(x, y);

			if ((x == 256) && (y == 0))
			{
				int stop = 0;
			}		

			hi = (int)GenerateHIndex();
			Q.push_back(hi);
		}

	sort(Q.begin(), Q.end());



	int x, y;
	i = 0;
	for (int i = 0; i < Q.size(); i++)
	{
		State = 0;
		CovertToXY(Q[i], x, y);
		M[x][y] = Y[i];
	}

}