#include "MatToVec.h"
#include "ImgToHilbert.h"
#include <vector>


MatToVec::MatToVec()
{
	W = H = 0;	
}

void MatToVec::SetMatrix(vector<vector<int>>& mat)
{
	M = mat;
	W = (int)M.size();
	H = (int)M[0].size();	
}

void MatToVec::SetVector(vector<int>& vec)
{
	Y = vec;
}



int MatToVec::IfOdd(int k)
{
	if (k % 2) return 1;
	return 0;
}


void MatToVec::InitTemporalM()
{
	if (T.size() != 0)
	{
		for (int i = 0; i < T.size(); i++)
			T[i].clear();
		T.clear();
	}

	T.resize(W);
	int k;
	for (k = 0; k < W; k++)
		T[k].resize(H);
}


/// 
/// Scan Line
///

void MatToVec::ScanLine()
{
	V.clear();
	V.resize(W * H);
	int x, y;
	int i = 0;
	for (y = 0; y < H; y++)
		for (x = 0; x < W; x++)
			V[i++] = M[x][y];
}


int MatToVec::DeScanLine()
{
	InitTemporalM();

	int x, y;
	int i = 0;
	for (y = 0; y < H; y++)
		for (x = 0; x < W; x++)
			T[x][y] = Y[i++];

	if (T == M) return 1;
	return 0;
}


/// 
/// Left Right
///


void MatToVec::LeftRight()
{
	V.clear();
	V.resize(W * H);

	int x, y;
	int i = 0;
	for (y = 0; y < H; y++)
	{
		if (IfOdd(y))
			for (x = W - 1; x >= 0; x--)
				V[i++] = M[x][y];
		else
			for (x = 0; x < W; x++)
				V[i++] = M[x][y];
	}
}

int MatToVec::DeLeftRight()
{
	InitTemporalM();

	int x, y;
	int i = 0;

	for (y = 0; y < H; y++)
	{
		if (IfOdd(y))
			for (x = W - 1; x >= 0; x--)
				T[x][y] = Y[i++];
		else
			for (x = 0; x < W; x++)
				T[x][y] = Y[i++];
	}
	if (T == M) return 1;
	return 0;
}


/// 
/// ZigZak
///



void MatToVec::ZigZag()
{
	V.clear();
	V.resize(W * H);

	int x, y;
	int xs, ys;
	int i = 0;


	for (int k = 0; k < W + H - 1; k++)
	{
		
		if (IfOdd(k)) // e lih
		{ // down

			if (k < W)
			{
				xs = k;
				y = 0;
			}
			else
			{
				xs = W - 1;
				y = k - W + 1;
			}

			for (int x = xs; x >= 0; x--)
				if (y < H)
					V[i++] = M[x][y++];
		}
		else
		{ // up

			if (k < H)
			{
				x = 0; 
				ys = k;
			}
			else
			{
				x = k - H + 1;
				ys = H - 1;
			}

			for (int y = ys; y >= 0; y--)
				if (x < W)
					V[i++] = M[x++][y];
		}
	}
}


int MatToVec::DeZigZag()
{
	InitTemporalM();

	int x, y;
	int xs, ys;
	int i = 0;


	for (int k = 0; k < W + H - 1; k++)
	{

		if (IfOdd(k)) // e lih
		{ // down

			if (k < W)
			{
				xs = k;
				y = 0;
			}
			else
			{
				xs = W - 1;
				y = k - W + 1;
			}

			for (int x = xs; x >= 0; x--)
				if (y < H)
					T[x][y++] = Y[i++];
		}
		else
		{ // up

			if (k < H)
			{
				x = 0;
				ys = k;
			}
			else
			{
				x = k - H + 1;
				ys = H - 1;
			}

			for (int y = ys; y >= 0; y--)
				if (x < W)
					T[x++][y] = Y[i++];
		}
	}
	if (T == M) return 1;
	return 0;
}



/// 
/// Strips
///


void MatToVec::Strips(int StripWidth)
{
	V.clear();
	V.resize(W * H);

	int x, y;
	y = 0;
	int y1;
	int sw = 0;
	int i = 0;
	while (y + StripWidth < H)
	{
		if (IfOdd(sw))
		{
			sw = 1;
			for (x = W - 1; x >= 0; x--)
				for (y1 = y; y1 < y + StripWidth; y1++)
					V[i++] = M[x][y1];
		}
		else
		{
			sw = 0;
			for (x = 0; x < W; x++)
				for (y1 = y; y1 < y + StripWidth; y1++)
					V[i++] = M[x][y1];
		}
		y += StripWidth;
	}

	for (y1 = y; y1 < H; y1++)
	{
		for (x = 0; x < W; x++)
			V[i++] =M[x][y1];
	}

}



int MatToVec::DeStrips(int StripWidth)
{
	InitTemporalM();

	int x, y;
	y = 0;
	int y1;
	int sw = 0;
	int i = 0;

	while (y + StripWidth < H)
	{
		if (IfOdd(sw))
		{
			sw = 1;
			for (x = W - 1; x >= 0; x--)
				for (y1 = y; y1 < y + StripWidth; y1++)
					T[x][y1] = Y[i++];
		}
		else
		{
			sw = 0;
			for (x = 0; x < W; x++)
				for (y1 = y; y1 < y + StripWidth; y1++)
					T[x][y1] = Y[i++];
		}
		y += StripWidth;
	}

	for (y1 = y; y1 < H; y1++)
	{
		for (x = 0; x < W; x++)
			T[x][y1]= Y[i++];
	}

	if (T == M) return 1;
	return 0;
}



/// 
/// Hilbert
///


void MatToVec::Hilbert()
{
	ImgToHilbert IToH;

	IToH.SetImage(W, H, M);
	IToH.ToHilbert();
	V = IToH.ReturnHilbertPixels();
}


int MatToVec::DeHilbert()
{
	InitTemporalM();

	ImgToHilbert IToHDe;

	IToHDe.DeSetImageResolution(W, H);
	IToHDe.DeReturnMatrix(Y, T);

	for (int x = 0; x < W; x++)
		for (int y = 0; y < H; y++)
		{
			if (M[x][y] != T[x][y])
			{
				int stop = 1;
			}

		}


	if (T == M) return 1;
	return 0;
}

