#include "MwI.h"
#include "AuxCompFunctions.h"

#pragma warning (disable : 4996)

MwI::MwI()
// l and u are the upper and lower value of in the data;
// this values are shiftedn on interval [0,u-l]
{
	ABCSize = lowerBound = upperBound = 0;
	ABC = ABCTail = NULL;
	Pool = NULL;
	O.clear();

}

MwI::~MwI()
{
	 Pool = DeleteList(Pool);
	 ABC = DeleteList(ABC);
}


ABCElement* MwI::DeleteList(ABCElement* List)
{
	if (List != NULL)
	{
		ABCElement *L, *D;
		L = List;
		while (L != NULL)
		{
			D = L;
			L = L->next;
			delete D;
		}
	}
	return NULL;
}

void MwI::AddToPool(ABCElement* el)
{
	el->next = Pool;
	Pool = el;

}

ABCElement* MwI::GetFromPool()
{
	ABCElement* p = Pool;
	Pool = Pool->next;
	return p;
}

int  MwI::CountPool()
{
	ABCElement* p = Pool;
	int i = 0;
	while (p != NULL)
	{
		i++;
		p = p->next;
	}
	return i;
}


void MwI::Init(int k, int l, int h, int first)
{
	lowerBound = l;
	upperBound = h;
	ABCSize = h - l + 1;
	delta = k;
	FirstEl = first;

	O.clear();
	ABCElement* abc;
	for (int i = 0; i < 2*ABCSize; i++)
	{
		abc = new(ABCElement);
		abc->el = 0;
		AddToPool(abc);		

	}

	InitABC();
//	PrintABC();
}


void MwI::PrintABC()
{
	ABCElement* L = ABC;
	printf("first %d delta %d low %d upper %d\n\n", FirstEl, delta, lowerBound, upperBound);
	while (L != NULL)
	{
		printf("%d\n", L->el);
		L = L->next;
	}
}

void MwI::PrintABC(int e)
{
	printf("----------------------");
	printf("element %d \n", e);
	ABCElement* L = ABC;
	while (L != NULL)
	{
		printf("%d\n", L->el);
		L = L->next;
	}
}

void MwI::PrintList(ABCElement* T)
{
	printf("----------------------");

	ABCElement* L = T;
	while (L != NULL)
	{
		printf("%d\n", L->el);
		L = L->next;
	}
}

void MwI::InitABC()
{

	int lBFlag = 1;
	int uBFlag = 1;
	int ABCIndex = 0;

	ABCElement* p = GetFromPool();
	p->el = FirstEl;
	p->next = NULL;
	ABC = ABCTail = p;


	int u, l;
	u = FirstEl + 1;
	l = FirstEl - 1;
	ABCIndex++;
	while ((lBFlag) || (uBFlag))
	{
		if (uBFlag)
		{
			p = GetFromPool();
			p->el = u++;
			p->next = NULL;
			ABCTail->next = p;
			ABCTail = p;

			if (u > upperBound)
				uBFlag = 0;
		}

		if (lBFlag)
		{
			p = GetFromPool();
			p->el = l--;
			p->next = NULL;
			ABCTail->next = p;
			ABCTail = p;

			if (l < lowerBound)
				lBFlag = 0;
		}
	}
}



int MwI::GetPosition(int v, ABCElement** qPrev, ABCElement** q)
{
	if (v == ABC->el)
		return 0;

	int i = 1;
	*qPrev = ABC;
	*q = (*qPrev)->next;

	while ((*q)->el != v)
	{
		i++;
		*qPrev = *q;
		*q = (*q)->next;
	}

	return i;
}


int MwI::GetValue(int position, ABCElement** qPrev, ABCElement** q)
{
	if (position == 0)
		return ABC->el;

	*qPrev = ABC;
	*q = (*qPrev)->next;
	int i = 1;

	while (i != position)
	{
		i++;
		*qPrev = *q;
		*q = (*q)->next;
	}
	return (*q)->el;
}



void MwI::MoveTToFront(ABCElement* head, ABCElement* tail, int a, int b)
{
	// first remove the elements from interval [a, b] from ABC 

//	PrintList(head);

	ABCElement *p1, *p2;
	int elToRemove = b - a + 1;
	int elindex = 0;


	while ((ABC != NULL) && (elindex < elToRemove) && (ABC->el >= a) && (ABC->el <= b))
	{	// first element should be removed
		p1 = ABC;
		ABC = ABC->next;
		AddToPool(p1);
		elindex++;
	}

	if (ABC != NULL)
	{ // elements in between should be removed

		p1 = ABC;
		p2 = p1->next;

		while ((p2 != NULL) && (elindex < elToRemove))
		{
			if ((p2->el < a) || (p2->el > b)) // out of range, the element remains
			{
				p1 = p2;
				p2 = p2->next;
			}
			else // remove the element
			{
				p1->next = p2->next;
				AddToPool(p2);
				p2 = p1->next;
				elindex++;
			}
		}
	}
	// now append the lest T to ABC

	tail->next = ABC;
	ABC = head;
}


vector<int> MwI::DoMwI(vector<int> I)
{

	O.resize(I.size()-1);
	int Oindex = 0;

	int i, j, position;
	int uv, lv;
	
	ABCElement *T, *Ttail;
	ABCElement* p;
	ABCElement *qPrev, *q;


	for (i = 1; i < I.size(); i++) // the value of the first element is known so we start at 1.
	{
		position = GetPosition(I[i], &qPrev, &q);
		O[Oindex++] = position;

		if (position != 0)
		{

			if (position > delta)
			{
				p = GetFromPool();
				p->el = I[i];
				p->next = NULL;
				T = Ttail = p;


				uv = I[i] + 1;
				lv = I[i] - 1;
				for (j = 0; j < delta; j++)
				{
					if (uv <= upperBound)
					{
						p = GetFromPool();
						p->el = uv++;
						p->next = NULL;
						Ttail->next = p;
						Ttail = p;
					}

					if (lv >= lowerBound)
					{
						p = GetFromPool();
						p->el = lv--;
						p->next = NULL;
						Ttail->next = p;
						Ttail = p;
					}
				}

				MoveTToFront(T, Ttail, lv+1, uv-1);
			}
			else // move element to front
			{
				qPrev->next = q->next;
				q->next = ABC;
				ABC = q;
			}

		}
	}


	return O;
}


vector<int> MwI::DoDeMwI(vector<int> I)
{
	O.clear();
	O.resize(I.size()+1);

	int Oindex = 0;

	int i, j, position;
	int value;
	int uv, lv;

	ABCElement* T, * Ttail;
	ABCElement* p;
	ABCElement* qPrev, * q;


	O[Oindex++] = FirstEl;

	
	for (i = 0; i < I.size(); i++)
	{
		position = I[i];
		value = GetValue(position, &qPrev, &q);
		O[Oindex++] = value;
		
		if (position != 0)
		{

			if (position > delta)
			{
				p = GetFromPool();
				p->el = value;
				p->next = NULL;
				T = Ttail = p;


				uv = value + 1;
				lv = value - 1;
				for (j = 0; j < delta; j++)
				{
					if (uv <= upperBound)
					{
						p = GetFromPool();
						p->el = uv++;
						p->next = NULL;
						Ttail->next = p;
						Ttail = p;
					}

					if (lv >= lowerBound)
					{
						p = GetFromPool();
						p->el = lv--;
						p->next = NULL;
						Ttail->next = p;
						Ttail = p;
					}
				}

				MoveTToFront(T, Ttail, lv + 1, uv - 1);
			}
			else // move element to front
			{
				qPrev->next = q->next;
				q->next = ABC;
				ABC = q;
			}
		}
	
	}
	return O;
}
