#include <stdio.h>
#include <string.h>
#include <vector>


#include "MyRawImageData.h"
#include "AuxCompFunctions.h"
#include "MTFWithIntegers.h"
#include "InverseFrequencies.h"
#include "UkkonenBwt.h"

#include "MatToVec.h"
#include "MwI.h"
#include "MyCPUTime.h"

#pragma warning (disable : 4996)

using namespace std;


double HBatchSum = 0;

void DetermineRange(int& lowerBound, int& upperBound, vector<int> I)
{
	lowerBound = 100000;
	upperBound = 0;

	for (int i = 0; i < I.size(); i++)
	{
		if (I[i] > upperBound)
			upperBound = I[i];
		else
			if (I[i] < lowerBound)
				lowerBound = I[i];
	}
}

void TransforWithMTF(vector<int>& V)
{
	
	MyCPUTime t1;
	t1.Start();

	int l, h;
	DetermineRange(l, h, V);

	MTFWithIntegers MTF;
	MTF.Init(l, h);

	vector<int> MTFI = MTF.DoMTF(V);

	t1.Stop();
	double d = t1.GetEllapsedCPUTime();
	printf("CPU Time %f\n", d);

	double HMTF = CalculateEntropy(DoVectorStatistic(MTFI));
	printf("\n ------------- MTF ---------------\n");
	printf(" H MTF: %f \n", HMTF);

	MTFWithIntegers DeMTF;
	DeMTF.Init(l, h);
	vector<int> DeMTFO = DeMTF.DoDeMTF(MTFI);
	if (V == DeMTFO)
		printf("DeMTF OK.");
	else
		printf("------> DeMTF failed.");

}




void TransforWithBWTandMTF(vector<int>& VI)
{
	MyCPUTime t1;
	t1.Start();

	int BWTIndex;
	vector<int> VBWT = BwtEncode(VI, BWTIndex);

	int l, h;
	DetermineRange(l, h, VBWT);

	MTFWithIntegers MTF;
	MTF.Init(l, h);

	vector<int> MTFI = MTF.DoMTF(VBWT);

	t1.Stop();
	double d = t1.GetEllapsedCPUTime();
	printf("CPU Time %f\n", d);


	double HMTF = CalculateEntropy(DoVectorStatistic(MTFI));
	printf("\n ------------- BWT + MTF ---------------\n");
	printf(" H MTF: %f \n", HMTF);


	MTFWithIntegers DeMTF;
	DeMTF.Init(l, h);
	vector<int> DeMTFO = DeMTF.DoDeMTF(MTFI);
	if (VBWT == DeMTFO)
		printf("DeMTF OK.");
	else
		printf("------> DeMTF failed.");

	vector<int> DeBWT = BwtDecode(VBWT, BWTIndex);

	if (VI == DeBWT)
		printf("\n DeBWT OK");
	else
		printf("------> DeBWT failed.");
}


/*
void TransforWith_k_MTF(vector<int>& V, int K_range)
{
	KMTFInt  KMTF;
	int first = V[0];
	int l, h;

	DetermineRange(l, h, V);

	KMTF.Init(K_range, l, h, first);

	vector<int> VKMTF = KMTF.DoKMTFInt(V);
	double HkMTF = CalculateEntropy(DoVectorStatistic(VKMTF));
	printf("\n ------------- k-MTF ---------------\n");
	printf(" H MTF: %f \n", HkMTF);

	KMTFInt DeKMTF;
	DeKMTF.Init(K_range, l, h, first);
	vector<int> DeKMTFO = DeKMTF.DoDeKMTFInt(VKMTF);
	if (V == DeKMTFO)
		printf(" DeMTF OK.");
	else
		printf(" ------> DeMTF failed.");
}
*/

void TransformWithMwI(vector<int>& V, int Delta)
{

	MyCPUTime t1;
	t1.Start();

	int first = V[0];
	int l, h;

	DetermineRange(l, h, V);
	MwI mvi;
	mvi.Init(Delta, l, h, first);
	vector<int> VMwI = mvi.DoMwI(V);
	
	t1.Stop();
	double d = t1.GetEllapsedCPUTime();
	printf("CPU Time %f\n", d);
	
	double H = CalculateEntropy(DoVectorStatistic(VMwI));
	printf("\n ------------- MwI ---------------\n");
	printf(" H MwI: %f \n", H);


	MwI DeMwI;
	DeMwI.Init(Delta, l, h, first);

	vector<int> DeMwIV = DeMwI.DoDeMwI(VMwI);
	if (V == DeMwIV)
		printf(" DeMwI OK.");
	else
	{
		printf(" ------> DeMwI failed.");
		int i = 0;
		while ((i < DeMwIV.size()) && (DeMwIV[i] == V[i]))
			i++;
		printf("-----> Error at index %d \n", i);

	}
}


void TransformWithBWTMwI(vector<int>& V, int delta)
{

	MyCPUTime t1;
	t1.Start();

	int BWTIndex;
	vector<int> VBWT = BwtEncode(V, BWTIndex);

	MwI  MWI;
	int first = VBWT[0];
	int l, h;

	DetermineRange(l, h, VBWT);
	MWI.Init(delta, l, h, first);
	vector<int> VMwI = MWI.DoMwI(VBWT);

	t1.Stop();
	double d = t1.GetEllapsedCPUTime();
	printf("CPU Time %f\n", d);

	double HkMTF = CalculateEntropy(DoVectorStatistic(VMwI));
	printf("\n ------------- BWT + MwI ---------------\n");
	printf(" H BWT+MwI: %f \n", HkMTF);

	MwI DeMWI;
	DeMWI.Init(delta, l, h, first);
	vector<int> DeMWIV = DeMWI.DoDeMwI(VMwI);

	vector<int> VDeBWT = BwtDecode(DeMWIV, BWTIndex);



	if (V == VDeBWT)
		printf(" De BWT + MwI OK.");
	else
		printf(" ------> De BWT + MwI failed.");

}


void TransformMwILoop(vector<int>& V)
{
	MwI  MWI;
	int first = V[0];
	int l, h;

	DetermineRange(l, h, V);

	int delta;
	double min = 1000;
	int MinD;
	for (delta = 0; delta <= 60; delta++)
	{
		MWI.Init(delta, l, h, first);

		vector<int> Vmwi = MWI.DoMwI(V);
		double H = CalculateEntropy(DoVectorStatistic(Vmwi));

		printf("delta %d  H MTF: %f \n", delta, H);

		if (min > H)
		{
			min = H;
			MinD = delta;
		}

	}
	printf("Min H %f, delta %d", min, MinD);
}



void TransformWithIFOriginal(vector<int>& V)
{
	MyCPUTime t1;
	t1.Start();

	InverseFrequencies IF;
	IF.InitOriginal(V);
	vector<int> VIF = IF.DoIFOriginal();

	t1.Stop();
	double d = t1.GetEllapsedCPUTime();
	printf("CPU Time %f\n", d);

	double HIF = CalculateEntropy(DoVectorStatistic(VIF));
	printf("\n ------------- IF Original ---------------\n");
	printf(" H MTF: %f \n", HIF);

	vector<ABCStruct> Abc = IF.ReturnAlphabet();

	InverseFrequencies DeIF;
	vector<int> DeIFV = IF.DeIFOriginal(Abc, VIF);

	if (DeIFV == V)
		printf(" DeIF Original OK.");
	else
		printf(" ------> DeIF Original failed.");
}




void TransformWithBWTIFOriginal(vector<int>& V)
{
	MyCPUTime t1;
	t1.Start();

	int BWTIndex;
	vector<int> VBWT = BwtEncode(V, BWTIndex);


	InverseFrequencies IF;
	IF.InitOriginal(VBWT);
	vector<int> VIF = IF.DoIFOriginal();
	t1.Stop();
	double d = t1.GetEllapsedCPUTime();
	printf("CPU Time %f\n", d);



	double HIF = CalculateEntropy(DoVectorStatistic(VIF));
	printf("\n ------------- BWT + IF Original ---------------\n");
	printf(" H MTF: %f \n", HIF);

	vector<ABCStruct> Abc = IF.ReturnAlphabet();

	InverseFrequencies DeIF;
	vector<int> DeIFV = IF.DeIFOriginal(Abc, VIF);

	if (DeIFV == VBWT)
		printf(" DeIF Original OK.");
	else
		printf(" ------> DeIF Original failed.");

	vector<int> DeBWT = BwtDecode(DeIFV, BWTIndex);

	if (V == DeBWT)
		printf("\n DeBWT OK");
	else
		printf("\n------> DeBWT failed.");
}




vector<vector<int>> ReturnTestMatrix(int w, int h)
{
	vector<vector<int>> M;
	M.resize(w);
	for (int i = 0; i < w; i++)
		M[i].resize(h);

	int k = 1;
	for (int x = 0; x < w; x++)
		for (int y = 0; y < h; y++)
			M[x][y] = k++;

	return M;
}


void TransformAll(vector<int> V, int Delta)
{

	double H = CalculateEntropy(DoVectorStatistic(V));
	printf("vhodna entropija: %f \n", H);

	int l, h;
	DetermineRange(l, h, V);
	MTFWithIntegers MTF;
	MTF.Init(l, h);
	vector<int> T = MTF.DoMTF(V);
	H = CalculateEntropy(DoVectorStatistic(T));
	printf("MTF: %f \n", H);

	InverseFrequencies IF;
	IF.InitOriginal(V);
	T.clear();
	T = IF.DoIFOriginal();
	H = CalculateEntropy(DoVectorStatistic(T));
	printf("IF: %f \n", H);

	MwI  MWI;
	int first = V[0];

	MWI.Init(Delta, l, h, first);
	T.clear();
	T = MWI.DoMwI(V);
	H = CalculateEntropy(DoVectorStatistic(T));
	printf("MTFwI: %f \n", H);

	int BWTIndex;
	vector<int> VBWT = BwtEncode(V, BWTIndex);

	DetermineRange(l, h, VBWT);
	MTFWithIntegers MTFBWT;
	MTFBWT.Init(l, h);
	T.clear();
	T = MTFBWT.DoMTF(VBWT);
	H = CalculateEntropy(DoVectorStatistic(T));
	printf("BWT + MTF: %f \n", H);

	InverseFrequencies IFB;
	IFB.InitOriginal(VBWT);
	T.clear();
	T = IFB.DoIFOriginal();

	H = CalculateEntropy(DoVectorStatistic(T));
	printf("BWT + IF: %f \n", H);

	MwI  MWI1;
	 first = V[0];

	MWI1.Init(Delta, l, h, first);
	T.clear();
	T = MWI1.DoMwI(VBWT);
	H = CalculateEntropy(DoVectorStatistic(T));
	printf("BWT + MwI: %f \n", H);


}
/////////////////// Batch

void TransforWithMTFBatch(vector<int>& V)
{
	int l, h;
	DetermineRange(l, h, V);

	MTFWithIntegers MTF;
	MTF.Init(l, h);

	vector<int> MTFI = MTF.DoMTF(V);
	double HMTF = CalculateEntropy(DoVectorStatistic(MTFI));
	printf("%0.3f \n", HMTF);
	HBatchSum += HMTF;

}

void TransformWithIFBatch(vector<int>& V)
{
	InverseFrequencies IF;
	IF.InitOriginal(V);
	vector<int> VIF = IF.DoIFOriginal();

	double HIF = CalculateEntropy(DoVectorStatistic(VIF));
	printf("%0.3f \n", HIF);
	HBatchSum += HIF;
}

void TransforWithMwIBatch(vector<int>& V, int delta)
{
	MwI  MWI;
	int first = V[0];
	int l, h;

	DetermineRange(l, h, V);

	MWI.Init(delta, l, h, first);

	vector<int> VMwI = MWI.DoMwI(V);
	double HMwI = CalculateEntropy(DoVectorStatistic(VMwI));
	printf("%0.3f \n", HMwI);
	HBatchSum += HMwI;

}

void TransforWithBWTandMTFBatch(vector<int>& VI)
{
	int BWTIndex;
	vector<int> VBWT = BwtEncode(VI, BWTIndex);

	int l, h;
	DetermineRange(l, h, VBWT);

	MTFWithIntegers MTF;
	MTF.Init(l, h);

	vector<int> MTFI = MTF.DoMTF(VBWT);
	double HMTF = CalculateEntropy(DoVectorStatistic(MTFI));
	printf("%0.3f \n", HMTF);
	HBatchSum += HMTF;
}


void TransformWithBWTIFBatch(vector<int>& V)
{
	int BWTIndex;
	vector<int> VBWT = BwtEncode(V, BWTIndex);
	InverseFrequencies IF;
	IF.InitOriginal(VBWT);
	vector<int> VIF = IF.DoIFOriginal();

	double HIF = CalculateEntropy(DoVectorStatistic(VIF));
	printf("%0.3f \n", HIF);
	HBatchSum += HIF;
}



void TransformWithBWTMwIBatch(vector<int>& V, int delta)
{

	int BWTIndex;
	vector<int> VBWT = BwtEncode(V, BWTIndex);


	MwI  MWI;
	int first = VBWT[0];
	int l, h;

	DetermineRange(l, h, VBWT);

	MWI.Init(delta, l, h, first);

	vector<int> VMWI = MWI.DoMwI(VBWT);
	double HMwI = CalculateEntropy(DoVectorStatistic(VMWI));
	printf("%0.3f \n", HMwI);
	HBatchSum += HMwI;
}



void ProcessBatch(MyRawImageData& MRID)
{

	printf("Organization of vector:\n");
	printf("1...Scan-line\n");
	printf("2...Left-right\n");
	printf("3...Zig-zag\n");
	printf("4...Strips\n");
	printf("5...Hilbert space filling curve\n");
	int ov = 1;
	scanf("%d", &ov);

	int sw= 0;
	if (ov == 4)
	{
		printf("strip width: ");
			scanf("%d", &sw);
	}

	printf("Transforms:\n");
	printf("1... MTF\n");
	printf("2... IF\n");
	printf("3... MwI\n");
	printf("4... BWT + MTF\n");
	printf("5... BWT + IF\n");
	printf("6... BWT + MwI\n");

	int trs;
	scanf("%d", &trs);

	int delta;
	if ((trs == 3) || (trs == 6))
	{
		printf("delta: ");
		scanf("%d", &delta);
	}



	vector<vector<int>>Image;
	vector<int> VI;
	MatToVec MtoV;

	while (MRID.ReturnNextBatchData(Image) == 1)
	{
		MtoV.SetMatrix(Image);
		switch (ov)
		{
		case 1: MtoV.ScanLine(); break;
		case 2: MtoV.LeftRight(); break;
		case 3: MtoV.ZigZag(); break;
		case 4:	MtoV.Strips(sw); break;
		case 5: MtoV.Hilbert(); break;
		default:
			break;
		}
		VI = MtoV.ReturnVector();


		switch (trs)
		{
		case 1: TransforWithMTFBatch(VI);  break;
		case 2: TransformWithIFBatch(VI);  break;
		case 3: TransforWithMwIBatch(VI, delta); break;
		case 4: TransforWithBWTandMTFBatch(VI); break;
		case 5: TransformWithBWTIFBatch(VI); break;
		case 6: TransformWithBWTMwIBatch(VI, delta); break;
		default:
			break;
		}
	}
	int NoOfBatches = MRID.ReturnNoOfBatches();
	printf("------\n");
	printf("Hsum: %f\n", HBatchSum);
	double average = HBatchSum / (double)NoOfBatches;
	printf("average H: %0.3f\n", average);

}


int main()
{
	MyRawImageData MRID;

	int ImageIndex = MRID.SelectImage();
	if (ImageIndex == 99) return 1;
	if (ImageIndex == 0)
	{
		ProcessBatch(MRID);
		return 1;
	}
	else
	{
		if (MRID.LoadBMP())
		{
			vector<vector<int>>Image = MRID.ReturnImageData();
			printf("Original H: %f \n", MRID.ReturnEntropy());

			printf("Organization of vector:\n");
			printf("1...Scan-line\n");
			printf("2...Left-right\n");
			printf("3...Zig-zag\n");
			printf("4...Strips\n");
			printf("5...Hilbert space filling curve\n");
			int ov = 1;
			scanf("%d", &ov);

			vector<int> VI;
			MatToVec MtoV;

			MtoV.SetMatrix(Image);
			int sw;

			switch (ov)
			{
			case 1: MtoV.ScanLine(); break;
			case 2: MtoV.LeftRight(); break;
			case 3: MtoV.ZigZag(); break;
			case 4:
				printf("strip width: ");
				scanf("%d", &sw);
				MtoV.Strips(sw);
				break;
			case 5: MtoV.Hilbert(); break;
			default:
				break;
			}

			VI = MtoV.ReturnVector();
			MtoV.SetVector(VI);

			double StDev = CalculateStandardDeviation(VI);
			printf("Standard deviation %f\n", StDev);

			int DeOvFlag = 0;
			switch (ov)
			{
			case 1: DeOvFlag = MtoV.DeScanLine(); break;
			case 2: DeOvFlag = MtoV.DeLeftRight(); break; 
			case 3: DeOvFlag = MtoV.DeZigZag(); break;
			case 4: DeOvFlag = MtoV.DeStrips(sw); break; 
			case 5: DeOvFlag = MtoV.DeHilbert(); break;
			default:
				break;
			}

			if (DeOvFlag)
				printf("->V to M is rigth. \n");
			else
				printf("-----> V to M failed. \n");

			printf("Transforms:\n");
			printf("1... MTF\n");
			printf("2... IF\n");
			printf("3... MwI\n");
			printf("4... BWT + MTF\n");
			printf("5... BWT + IF\n");
			printf("6... BWT + MTFwI\n");
			printf("7... ALL\n");
			printf("8... Determine MwI delta\n");

			int trs;
			scanf("%d", &trs);


			switch (trs)
			{
			case 1: TransforWithMTF(VI);  break;
			case 2: TransformWithIFOriginal(VI);  break;
			case 3: TransformWithMwI(VI, 11); break;
			case 4: TransforWithBWTandMTF(VI); break;
			case 5: TransformWithBWTIFOriginal(VI); break;
			case 6: TransformWithBWTMwI(VI, 12); break;
			case 7: TransformAll(VI, 12);  break;
			case 8: TransformMwILoop(VI); break;
			default:
				break;
			}

		}
	}
	printf("\n\n");

}